"""Test StructuredQuadGraph."""
from numpy.testing import assert_array_almost_equal, assert_array_equal

from landlab.graph import DualStructuredQuadGraph, DualUniformRectilinearGraph


def test_create():
    """Test creating a quad graph."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))

    assert graph.number_of_nodes == 9
    assert graph.number_of_links == 12
    assert graph.number_of_patches == 4

    assert graph.number_of_corners == 4
    assert graph.number_of_faces == 4
    assert graph.number_of_cells == 1


def test_create_raster():
    """Test creating a quad graph."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=(2., 3.))

    assert graph.number_of_nodes == 12
    assert graph.number_of_links == 17
    assert graph.number_of_patches == 6

    assert graph.number_of_corners == 6
    assert graph.number_of_faces == 7
    assert graph.number_of_cells == 2


def test_raster_spacing():
    """Test the spacing keyword for raster."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=(2., 3.))

    assert_array_equal(
        graph.length_of_link,
        [3., 3., 3., 2., 2., 2., 2., 3., 3., 3., 2., 2., 2., 2., 3., 3., 3.],
    )
    assert_array_equal(graph.length_of_face, [3., 3., 2., 2., 2., 3., 3.])


def test_raster_spacing_as_scalar():
    """Test the spacing keyword as a scalar for raster."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=2.)

    assert_array_equal(
        graph.length_of_link,
        [2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2., 2.],
    )
    assert_array_equal(graph.length_of_face, [2., 2., 2., 2., 2., 2., 2.])


def test_raster_origin():
    """Test the origin keyword for raster."""
    graph = DualUniformRectilinearGraph((3, 4), origin=(-1., 10.))

    assert_array_equal(
        graph.xy_of_node[:, 0],
        [10., 11., 12., 13., 10., 11., 12., 13., 10., 11., 12., 13.],
    )
    assert_array_equal(
        graph.xy_of_node[:, 1], [-1., -1., -1., -1., 0., 0., 0., 0., 1., 1., 1., 1.]
    )

    assert_array_equal(graph.xy_of_corner[:, 0], [10.5, 11.5, 12.5, 10.5, 11.5, 12.5])
    assert_array_equal(graph.xy_of_corner[:, 1], [-.5, -.5, -.5, .5, .5, .5])


def test_raster_origin_as_scalar():
    """Test the origin keyword as a scalar for raster."""
    graph = DualUniformRectilinearGraph((3, 4), origin=-1.)

    assert_array_equal(
        graph.xy_of_node[:, 0], [-1., 0., 1., 2., -1., 0., 1., 2., -1., 0., 1., 2.]
    )
    assert_array_equal(
        graph.xy_of_node[:, 1], [-1., -1., -1., -1., 0., 0., 0., 0., 1., 1., 1., 1.]
    )

    assert_array_equal(graph.xy_of_corner[:, 0], [-.5, .5, 1.5, -.5, .5, 1.5])
    assert_array_equal(graph.xy_of_corner[:, 1], [-.5, -.5, -.5, .5, .5, .5])


def test_perimeter_corners():
    """Test the perimeter corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.perimeter_corners, [1, 3, 2, 0])


def test_length_of_face():
    """Test length of faces."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))

    assert_array_almost_equal(graph.length_of_face, [1.5, 1.5, 1.5, 1.5])
    assert_array_almost_equal(graph.width_of_face, [1.5, 1.5, 1.5, 1.5])
    assert graph.length_of_face is graph.width_of_face


def test_area_of_cell():
    """Test areas of patches."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_almost_equal(graph.area_of_cell, [2.25])


def test_corners_at_cell():
    """Test corners of cells."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.corners_at_cell, [[3, 2, 0, 1]])


def test_cells_at_corner():
    """Test areas of patches."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(
        graph.cells_at_corner,
        [[0, -1, -1, -1], [-1, 0, -1, -1], [-1, -1, -1, 0], [-1, -1, 0, -1]],
    )


def test_cells_at_face():
    """Test cells on either side of faces."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.cells_at_face, [[-1, 0], [0, -1], [-1, 0], [0, -1]])


def test_faces_at_cell():
    """Test faces that form cells."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.faces_at_cell, [[2, 3, 1, 0]])


def test_corners_at_face():
    """Test corners at face tail and head."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.corners_at_face, [[0, 1], [0, 2], [1, 3], [2, 3]])


def test_faces_at_corner():
    """Test faces around corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(
        graph.faces_at_corner,
        [[0, 1, -1, -1], [-1, 2, 0, -1], [3, -1, -1, 1], [-1, -1, 3, 2]],
    )


def test_face_dirs_at_corner():
    """Test face directions at corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(
        graph.face_dirs_at_corner,
        [[-1, -1, 0, 0], [0, -1, 1, 0], [-1, 0, 0, 1], [0, 0, 1, 1]],
    )


def test_cell_at_node():
    """Test cell-node connectivity."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.cell_at_node, [-1, -1, -1, -1, 0, -1, -1, -1, -1])

    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(
        graph.cell_at_node, [-1, -1, -1, -1, -1, 0, 1, -1, -1, -1, -1, -1]
    )


def test_link_at_face():
    """Test link-face connectivity."""
    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(graph.link_at_face, [4, 5, 7, 8, 9, 11, 12])
    assert_array_equal(
        graph.face_at_link,
        [-1, -1, -1, -1, 0, 1, -1, 2, 3, 4, -1, 5, 6, -1, -1, -1, -1],
    )


def test_corner_at_face():
    """Test corner-face connectivity."""
    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(
        graph.corners_at_face, [[0, 1], [1, 2], [0, 3], [1, 4], [2, 5], [3, 4], [4, 5]]
    )
    assert_array_equal(graph.corner_at_face_tail, [0, 1, 0, 1, 2, 3, 4])
    assert_array_equal(graph.corner_at_face_head, [1, 2, 3, 4, 5, 4, 5])



import numpy as np


def add_halo(data, halo=1, halo_value=None):
    """Add a halo of no data value to data.

    Parameters
    ----------
    data : array-like
        Array to add the halo to.
    halo : int, optional
        The size of the halo.
    halo_value : float, optional
        Value to fill the halo with. If not provided, the new data will is not
        initialized.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.utils import add_halo
    >>> data = np.array([[1, 2, 3], [4, 5, 6]])
    >>> add_halo(data, halo_value=9)
    array([[9, 9, 9, 9, 9], [9, 1, 2, 3, 9], [9, 4, 5, 6, 9], [9, 9, 9, 9, 9]])
    """
    if halo < 0:
        raise ValueError("halo must be greater than or equal to zero")
    elif halo == 0:
        return data.copy()

    data_with_halo = np.empty([dim + 2 * halo for dim in data.shape], dtype=data.dtype)
    data_with_halo[halo:-halo, halo:-halo] = data
    if halo_value is not None:
        data_with_halo[:halo, :] = halo_value
        data_with_halo[-halo:, :] = halo_value
        data_with_halo[:, :halo] = halo_value
        data_with_halo[:, -halo:] = halo_value

    return data_with_halo



from .structured_quad import (
    StructuredQuadGraph,
    RectilinearGraph,
    UniformRectilinearGraph,
)
from .dual_structured_quad import (
    DualUniformRectilinearGraph,
    DualRectilinearGraph,
    DualStructuredQuadGraph,
)


__all__ = [
    "StructuredQuadGraph",
    "RectilinearGraph",
    "UniformRectilinearGraph",
    "DualUniformRectilinearGraph",
    "DualRectilinearGraph",
    "DualStructuredQuadGraph",
]



import os

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import create_grid

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

SIMPLE_PARAMS_STR = """
grid:
  RasterModelGrid:
    - [4, 5]
    - xy_spacing: [3, 4]
fields:
  at_node:
    topographic__elevation:
      plane:
        - point: [1, 1, 1]
          normal: [-2, -1, 1]
boundary_conditions:
  - set_closed_boundaries_at_grid_edges:
    - True
    - True
    - True
    - True
"""


def test_no_grid_value():
    dict_like = {"foo": "bar"}
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_grid_name():
    dict_like = {"grid": "MagicModelGrid"}
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_two_grid_types():
    dict_like = {"grid": {"RasterModelGrid": [(4, 5)], "HexModelGrid": [6, 7]}}
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_field_location():
    dict_like = {
        "grid": {"RasterModelGrid": [(4, 5)]},
        "fields": {"at_bad_location": {"foo": "bar"}},
    }
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_field_function():
    dict_like = {
        "grid": {"RasterModelGrid": [(4, 5)]},
        "fields": {"at_node": {"new_field_name": {"not_a_function": ["bar", "spam"]}}},
    }
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_simple_create(tmpdir):
    """Load parameters from YAML-formatted file."""
    with tmpdir.as_cwd():
        with open("params.yaml", "w") as fp:
            fp.write(SIMPLE_PARAMS_STR)

        mg = create_grid("./params.yaml")

    assert mg.number_of_nodes == 20
    assert "topographic__elevation" in mg.at_node

    x_of_node = np.array(
        [
            0.,
            3.,
            6.,
            9.,
            12.,
            0.,
            3.,
            6.,
            9.,
            12.,
            0.,
            3.,
            6.,
            9.,
            12.,
            0.,
            3.,
            6.,
            9.,
            12.,
        ]
    )

    status_at_node = np.array(
        [4, 4, 4, 4, 4, 4, 0, 0, 0, 4, 4, 0, 0, 0, 4, 4, 4, 4, 4, 4], dtype=np.uint8
    )
    topographic__elevation = np.array(
        [
            [-2., 4., 10., 16., 22.],
            [2., 8., 14., 20., 26.],
            [6., 12., 18., 24., 30.],
            [10., 16., 22., 28., 34.],
        ]
    )

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(
        topographic__elevation,
        np.round(mg.at_node["topographic__elevation"].reshape(mg.shape), decimals=2),
    )


def test_esri_ascii_create():
    filename = os.path.join(_TEST_DATA_DIR, "4_x_3_no_nodata_value.asc")
    dict_like = {
        "grid": {
            "RasterModelGrid": [(4, 3), {"xy_spacing": 10, "xy_of_lower_left": (1, 2)}]
        },
        "fields": {
            "at_node": {"topographic__elevation": {"read_esri_ascii": [filename]}}
        },
    }
    mg = create_grid(dict_like)

    x_of_node = np.array([1., 11., 21., 1., 11., 21., 1., 11., 21., 1., 11., 21.])
    y_of_node = np.array([2., 2., 2., 12., 12., 12., 22., 22., 22., 32., 32., 32.])
    status_at_node = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1], dtype=np.uint8)
    topographic__elevation = np.array(
        [9., 10., 11., 6., 7., 8., 3., 4., 5., 0., 1., 2.]
    )

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(mg.y_of_node, y_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(topographic__elevation, mg.at_node["topographic__elevation"])


def test_read_netcdf_create():
    filename = os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc")
    dict_like = {
        "grid": {"RasterModelGrid": [(4, 3)]},
        "fields": {"at_node": {"surface__elevation": {"read_netcdf": [filename]}}},
    }
    mg = create_grid(dict_like)
    x_of_node = np.array([0., 1., 2., 0., 1., 2., 0., 1., 2., 0., 1., 2.])
    y_of_node = np.array([0., 0., 0., 1., 1., 1., 2., 2., 2., 3., 3., 3.])
    status_at_node = np.array([1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1], dtype=np.uint8)
    surface__elevation = np.array([0., 1., 2., 3., 4., 5., 6., 7., 8., 9., 10., 11.])

    assert_array_equal(mg.x_of_node, x_of_node)
    assert_array_equal(mg.y_of_node, y_of_node)
    assert_array_equal(status_at_node, mg.status_at_node)
    assert_array_equal(surface__elevation, mg.at_node["surface__elevation"])


def test_two_boundary_condition_dicts():
    dict_like = {
        "grid": {"RasterModelGrid": [(4, 3)]},
        "boundary_conditions": [
            {
                "set_closed_boundaries_at_grid_edges": [True, True, True, True],
                "not_a_function": [True, False],
            }
        ],
    }
    with pytest.raises(ValueError):
        create_grid(dict_like)


def test_bad_boundary_condition_functions():
    filename = os.path.join(_TEST_DATA_DIR, "bad_boundary.yaml")
    with pytest.raises(ValueError):
        create_grid(filename)



"""
Modules that read/write ModelGrids from various file formats.
"""
from .shapefile import read_shapefile
from .esri_ascii import read_esri_ascii, read_asc_header, write_esri_ascii
from .esri_ascii import (
    MissingRequiredKeyError,
    KeyTypeError,
    KeyValueError,
    DataSizeError,
    BadHeaderLineError,
    MismatchGridDataSizeError,
    MismatchGridXYSpacing,
    MismatchGridXYLowerLeft,
)

__all__ = [
    "read_esri_ascii",
    "read_asc_header",
    "read_shapefile",
    "write_esri_ascii",
    "MissingRequiredKeyError",
    "KeyTypeError",
    "DataSizeError",
    "BadHeaderLineError",
    "KeyValueError",
    "MismatchGridDataSizeError",
    "MismatchGridXYSpacing",
    "MismatchGridXYLowerLeft",
]
